"""
Simulation environment modeling the stochastic energy consumption of an AEV
traversing the streets of Manhattan, NY
"""

import os
import ast
import json
import folium
import timeit
import numpy as np
import networkx as nx
from decimal import Decimal
from folium import plugins
from fimdp import energy_solver
from fimdp import consMDP
from energy_solver import BUCHI
import matplotlib.pyplot as plt
from matplotlib.ticker import PercentFormatter


"""
Function for parsing electric vehicle energy consumption data
from a json file. 
"""

def load_json(filename):
    with open(filename,'r') as f:
        return json.load(f)

def convert(g):
    states = []
    actions = dict()

    mdp = consMDP.ConsMDP()

    for node in g["nodes"]:
        if node["action"]:
            actions[node["label"]] = dict()
        else:
            states.append(node)

    for s in states:
        mdp.new_state(s["reload"], s["label"])

    for edge in g["edges"]:
        fr = edge["tail"]
        to = edge["head"]
        if to in actions:
            actions[to]["from"] = fr
            actions[to]["cons"] = edge["consumption"]
        else:
            dist = actions[fr].get("dist")
            to = mdp.state_with_name(to)
            if dist is None:
                actions[fr]["dist"] = dict()
            actions[fr]["dist"][to] = Decimal(f'{edge["probability"]}')

    for label, a in actions.items():
        fr = mdp.state_with_name(a["from"])
        mdp.add_action(fr, a["dist"], label, a["cons"])

    return mdp


def get_target_set(g, mdp):
    T = set()
    for t in g["T"]:
        T.add(mdp.state_with_name(t["label"]))
    return T

def parse(filename):
    js = load_json(filename)
    
    mdp = convert(js)
    target = get_target_set(js, mdp)
    
    return (mdp, target)


def timeit_difftargets(m, cap, target_size = 100, num_samples = 100, num_tests=5, obj=BUCHI):

    """Returns a list of compute times for given objective for the NYC AEV
    problem with varying target sets.

    Parameters
    ----------
    m : mdp;  object of class ConsMDP.
       A valid Markov Decision Process.

    cap : positive integer
       The energy capacity of the agent.

    target_size : positive integer, optional
       Size of the set of random targets to be generated for calculating the
       compute time variation. Takes a default value of 100.

    num_samples : positive integer, optional
       Number of times a random target set is generated and it's compute time
       for any given objective is calculated. Takes a default of 100.

    num_tests : positive integer, optional
       Number of times the compute time is calculated using timeit for each
       individual test. Takes a default value of 5.

    obj : one of MIN_INIT_CONS, SAFE, POS_REACH, AS_REACH, BUCHI, optional
       Objective for which the compute times are calculated. Takes BUCHI as
       the default value.

    Returns
    -------
    comptime : array_like
       List of computational times for different target sets.

    Examples
    --------
    >>> m, targets = ch_parser.parse('NYCstreetnetwork.json')
    >>> comptime = timeit_difftargets(m, cap=200, target_size=100)

    Notes
    -----
    The values of timings returned might significantly vary depending on the
    machine configuration and the target states set.

    """

    comptime = np.empty(num_samples)
    for i in range(num_samples):
        targets = np.random.randint(0, high=m.num_states, size=(target_size))
        def calc_time():
            s = energy_solver.EnergySolver(m, cap=cap, targets=targets)
            s.get_strategy(obj, recompute=True)
        comptime[i] = timeit.timeit(calc_time, number=num_tests)/num_tests
    return comptime


def timeit_diffcaps(m, targets, cap_bound , num_samples = 20, num_tests=10, obj=BUCHI):

    """Returns a list of tuples where each tuple consists of the energy capacity
    and its corresponding computational time for a given objective and target
    set in the NYC AEV problem.

    Parameters
    ----------
    m : mdp;  object of class ConsMDP
       A valid Markov Decision Process.

    targets : array_like, list of target states
       A list containing all the target states.

    cap_bound : positive integer
       Upper bound of the interval specifying the allowed energy capacity
       for the agent.

    num_samples : positive integer, optional
       Number of equally spaced samples to be collected from the interval
       [0, cap_bound] for evaluation of their computational time. Takes a
       default value of 20.

    num_tests : positive integer, optional
       Number of times the compute time is calculated using timeit for each
       sampled capacity value. Takes a default value of 10.

    obj : one of MIN_INIT_CONS, SAFE, POS_REACH, AS_REACH, BUCHI, optional
       Objective for which the compute times are calculated. Takes BUCHI as
       the default value.

    Returns
    -------
    comptime : array_like
       List with tuples of capacity and expected computational time for a fixed
       set of targets.

    Examples
    --------
    >>> m, targets = ch_parser.parse('NYCstreetnetwork.json')
    >>> comptime = timeit_diffcaps(m, targets=targets, cap_bound=200, obj=BUCHI)

    Notes
    -----
    The values of timings returned might significantly vary depending on the
    machine configuration.

    """

    comptime = []

    cap_list = np.linspace(1,cap_bound, num_samples,dtype = int)
    for i in range(num_samples):
        def calc_time():
            s = energy_solver.EnergySolver(m, cap=cap_list[i], targets=targets)
            s.get_strategy(obj, recompute=True)
        comptime.append((cap_list[i], timeit.timeit(calc_time, number=num_tests)/num_tests))
    return comptime

def timeit_difftargetsizes(m, cap, size_bound , num_samples = 20, num_tests=10, obj=BUCHI):

    """Returns a list of tuples where each tuple consists of the target size
        and its corresponding computational time for a given objective and capacity
        in the NYC AEV problem.

    Parameters
    ----------
    m : mdp;  object of class ConsMDP
       A valid Markov Decision Process.

    cap : positive integer.
       The energy capacity of the agent.

    size_bound : positive integer
       Upper bound of the interval specifying the allowed size of randomly
       generated target set for the agent.

    num_samples : positive integer, optional
       Number of equally spaced samples to be collected from the interval
       [20, size_bound] for evaluation of their computational time. Takes a
       default value of 20.

    num_tests : positive integer, optional
       Number of times the compute time is calculated using timeit for each
       target set. Takes a default value of 10.

    obj : one of MIN_INIT_CONS, SAFE, POS_REACH, AS_REACH, BUCHI, optional
       Objective for which the compute times are calculated. Takes BUCHI as
       the default value.

    Returns
    -------
    comptime : array_like
       List with tuples of target set size and expected computational time for a fixed
       capacity.

    Examples
    --------
    >>> m, targets = ch_parser.parse('NYCstreetnetwork.json')
    >>> comptime = timeit_difftargetsizes(m, cap=100, size_bound=200, obj=BUCHI)

    Notes
    -----
    The values of timings returned might significantly vary depending on the
    machine configuration.

    """

    comptime = []

    targetsize_list = np.linspace(20,size_bound, num_samples, dtype = int)
    targetsize_list = targetsize_list.astype(np.int64)
    for i in range(num_samples):
        def calc_time():
            targets = np.random.randint(0, high=m.num_states, size=(targetsize_list[i]))
            s = energy_solver.EnergySolver(m, cap=cap, targets=targets)
            s.get_strategy(obj, recompute=True)
        comptime.append((targetsize_list[i], timeit.timeit(calc_time, number=num_tests)/num_tests))
    return comptime

def histogram(comptime):
    """
    PLot histogram of the computation time for different targets set of same
    size. Y axis is expressed in terms of pecentage of samples. 
    """

    plt.hist(comptime, weights=np.ones(len(comptime))/len(comptime),color='skyblue')
    plt.xlabel('Computation Time (sec)')
    plt.title('Histogram of Buchi Computational Time. Mean: {} and SD: {}'.format(round(np.mean(comptime),4), round(np.std(comptime),4)))
    plt.gca().yaxis.set_major_formatter(PercentFormatter(1))
    plt.show()

def visualize_strategy(strategy, m, targets):
    """
    Visualize the actions suggested by a strategy for a given MDP modeling an
    AEV in a street network.
    
    Parameters
    ----------
    
    strategy : list
        A list of length equal to the number of states where each element is a
        dict element with keys are energy values and values as the action.
    
    m : mdp;  object of class ConsMDP
       A valid Markov Decision Process.
    
    targets : array_like, list of target states
       A list containing all the target states.
       
    Returns
    -------
    mapobj; folium map object
        A folium map object containing all the states and actions.
    
    Examples
    --------
    >>> m, targets = ch_parser.parse('NYCstreetnetwork.json')
    >>> s = energy_solver.EnergySolver(m, cap=100, targets=targets)
    >>> strategy = s.get_strategy(BUCHI, recompute=True)
    >>> mapobj = visualize_strategy(strategy, m, targets)
    
    """
    # Initialize
    states_nostrategy = [] 
    states_reload = [] 
    states_target = []
    strategy_updated = {}
    map_statelabels = m.names
    
    # Map states to original labels
    for index in range(m.num_states):
        if not map_statelabels[index][:2] == 'ps':
            if strategy[index]:
                strategy_updated.update({map_statelabels[index]: strategy[index]})
            else:
                states_nostrategy.append(map_statelabels[index])

    # Map target and reload states to original labels
    for item in targets:
        states_target.append(map_statelabels[item])
    for index, item in enumerate(m.reloads):
        if item:
            states_reload.append(map_statelabels[index])
            
    # Extract resultant states for actions
    dynamics = {}
    with open('NYCstreetnetwork.json','r') as f:
        raw_data = json.load(f)
        for edge in raw_data["edges"]:
            tail = edge["tail"]
            head = edge["head"]
            if tail[:2] == 'pa':
                action_label = tail[4:]
                dynamics.update({action_label:head})
                
    # Map actions in strategy to resultant states
    for key, value in strategy_updated.items():
        for energy, action_label in value.items():
            value.update({energy: dynamics[action_label]})
    
    mapobj = create_geoplot(states_nostrategy, states_reload, states_target, strategy_updated)   
    return mapobj         

def create_geoplot(states_nostrategy, states_reload, states_target, strategy_updated):
    
    """
    Helper function to visualize the data from MDP and strategy on an OpenStreetMap.
    Uses geodata of the street network stored as a graph in 'nyc.graphml' file.
    """
    
    # Load NYC Geodata
    path = os.path.abspath("nyc.graphml")
    G = nx.MultiDiGraph(nx.read_graphml(path))
    for _, _, data in G.edges(data=True, keys=False):
        data['speed_mean'] = float(data['speed_mean'])
        data['speed_sd'] = float(data['speed_sd'])
        data['time_mean'] = float(data['time_mean'])
        data['time_sd'] = float(data['time_sd'])
        data['energy_levels'] = ast.literal_eval(data['energy_levels'])
    for _, data in G.nodes(data=True):
        data['reload'] = ast.literal_eval(data['reload'])
        data['lat'] = float(data['lat'])
        data['lon'] = float(data['lon'])

    # Create baseline map with edges 
    nodes_all = {}
    for node in G.nodes.data():
        name = str(node[0])
        point = [node[1]['lat'], node[1]['lon']]
        nodes_all[name] = point
    global_lat = []; global_lon = []
    for name, point in nodes_all.items():
        global_lat.append(point[0])
        global_lon.append(point[1])
    min_point = [min(global_lat), min(global_lon)]
    max_point =[max(global_lat), max(global_lon)]
    m = folium.Map(zoom_start=1, tiles='cartodbpositron')
    m.fit_bounds([min_point, max_point])
    for edge in G.edges:
        points = [(G.nodes[edge[0]]['lat'], G.nodes[edge[0]]['lon']),
                  (G.nodes[edge[1]]['lat'], G.nodes[edge[1]]['lon'])]
        folium.PolyLine(locations=points,
                        color='gray',
                        weight=2,
                        opacity=0.8).add_to(m)
    
    for key, value in strategy_updated.items():
        color = '#2f2f2f'
        for energy, end_state in value.items():
            points = [(G.nodes[key]['lat'], G.nodes[key]['lon']),
                      (G.nodes[end_state]['lat'], G.nodes[end_state]['lon'])]
            line = folium.PolyLine(locations=points,
                                   color=color,
                                   tooltip=str(energy),
                                   weight=1.5).add_to(m)
            attr = {'fill': color, 'font-size': '12'}
            plugins.PolyLineTextPath(line,'\u25BA',
                                     repeat=False,
                                     center=True,
                                     offset=3.5,
                                     attributes=attr).add_to(m)
            folium.CircleMarker(location=[G.nodes[key]['lat'], G.nodes[key]['lon']],
                        radius= 2,
                        color=color,
                        fill=True).add_to(m)
    
    # Add reload states, target states, and states with no prescribed action
    nodes_reload = {}
    nodes_target = {}
    nodes_nostrategy = {}
    for node in G.nodes.data():
        if node[0] in states_reload:
            name = str(node[0])
            point = [node[1]['lat'], node[1]['lon']]
            nodes_reload[name] = point
        if node[0] in states_target:
            name = str(node[0])
            point = [node[1]['lat'], node[1]['lon']]
            nodes_target[name] = point
        if node[0] in states_nostrategy:
            name = str(node[0])
            point = [node[1]['lat'], node[1]['lon']]
            nodes_nostrategy[name] = point

    # Plot reload states
    for node_name, node_point in nodes_reload.items():
        folium.CircleMarker(location=[node_point[0], node_point[1]],
                        radius= 3,
                        popup = 'reload state',
                        color="#22af4b",
                        fill_color = "#22af4b",
                        fill_opacity=1,
                        fill=True).add_to(m)
    # Plot target nodes
    for node_name, node_point in nodes_target.items():
        folium.CircleMarker(location=[node_point[0], node_point[1]],
                        radius= 3,
                        popup = 'target state',
                        color="#0f89ca",
                        fill_color = "#0f89ca",
                        fill_opacity=1,
                        fill=True).add_to(m)
    # Plot no strategy nodes
    for node_name, node_point in nodes_nostrategy.items():
        folium.CircleMarker(location=[node_point[0], node_point[1]],
                        radius= 3,
                        popup = 'no guarantees state',
                        color='red',
                        fill_color = 'red',
                        fill_opacity=1,
                        fill=True).add_to(m)
    return m







__version__ = "0.0.4"


def setup():
    """
    Function to setup the visualization environment in notebooks
    """
    
    from matplotlib import animation, rc
    rc('animation', html='jshtml')
    


"""
Gridworld modeling motion of a constant horizontal velocity UUV in waters
with ocean currents.
"""

# import required packages
import ast
import numpy as np
from fimdp import consMDP
from decimal import Decimal
from matplotlib import pyplot as plt
from scipy.stats import norm
import matplotlib.animation as animation


# Define colors
# 0: light blue; 1 : light gray; 2 : dark gray; 3 : green; 4 : orange; 5: dark blue
COLORS = {0:np.array([0.85,1.0,1.0]), 1:np.array([0.54,0.54,0.54]), \
          2:np.array([0.42,0.42,0.42]), 3:np.array([0.0,1.0,0.0]), \
          4:np.array([1.0,0.37,0.008]), 5:np.array([0.0,0.0,1.0])}


class SingleAgentEnv:

    def __init__(self, grid_size, agent_capacity, reload_list, target_list, init_state=None, agent_velocity=5, heading_sd=0.524, weakaction_cost=1, strongaction_cost=2):
        """Class that models a Markov Decision Process where the agent and the environment model the dynamics of a UUV and 
        the ocean currents it is operating in.  

        :param grid_size: tuple (m,n) of length two with positive integers that denotes the dimensions of the 2D grid to be generated. 
        :type grid_size: int
        :param agent_capacity: Non negative number denoting the energy capacity of the agent
        :type agent_capacity: float
        :param waction_cost: Positive number that denotes the energy consumed by the agent for any weak action.
        :type waction_cost: float
        :param saction_cost: Positive number that denotes the energy consumed by the agent for any strong action.
        :type saction_cost: float
        :param init_state: An integer denoting the initial state of the agent, defaults to None
        :type init_state: int, optional
        """

        # user inputs and other required constants
        self.grid_size = grid_size
        self.agent_capacity = agent_capacity
        self.waction_cost = weakaction_cost
        self.saction_cost = strongaction_cost
        self.reload_list = reload_list
        self.target_list = target_list
        self.init_state = init_state
        self.position = None
        self.target_state = None
        self.stategy = None
        self.states_history = []
        self.actions_history = []
        self.targets_history = []
        self.time_steps = 0
        self.consmdp = None
        

        # initialize required variables
        self.num_cells = grid_size[0]*grid_size[1]
        self.num_actions = 8
        self.is_reload = np.zeros(self.num_cells)
        self.agent_energy = self.agent_capacity
        self.agent_P = np.zeros(
            [self.num_cells, self.num_actions, self.num_cells])

        # initialize and generate current flow field -- DEMO CONFIG ONLY
        self.AGENT_V = agent_velocity
        self.HEADING_SD = heading_sd
        self.FLOW_VX = 0.1*self.AGENT_V * \
            np.ones([self.num_cells, self.num_cells])
        self.FLOW_VY = 0.2*self.AGENT_V * \
            np.ones([self.num_cells, self.num_cells])
        self.FLOW_MAG = np.hypot(self.FLOW_VX, self.FLOW_VY)
        self.FLOW_THETA = np.arctan2(self.FLOW_VY, self.FLOW_VX)
        self.THETA_LIST = [np.pi/2, -np.pi/2, 0, np.pi]

        # generate state transition probabilities
        for i in range(self.num_cells):
            for j in range(self.num_actions):

                # north, strong north - action 0 and action 4
                if j in [0, 4]:
                    if i - self.grid_size[1] < 0:
                        self.agent_P[i, j, i] = Decimal(str(1))
                    else:
                        self.agent_P[i, j, i-self.grid_size[1]
                                     ] = Decimal(str(1))

                # south, strong south - action 1 and action 5
                elif j in [1, 5]:
                    if i + self.grid_size[1] >= self.num_cells:
                        self.agent_P[i, j, i] = Decimal(str(1))
                    else:
                        self.agent_P[i, j, i+self.grid_size[1]
                                     ] = Decimal(str(1))

                # east, strong east - action 2 and action 6
                elif j in [2, 6]:
                    if i % self.grid_size[1] == self.grid_size[1]-1:
                        self.agent_P[i, j, i] = Decimal(str(1))
                    else:
                        self.agent_P[i, j, i+1] = Decimal(str(1))

                # west, strong west - action 3 and action 7
                elif j in [3, 7]:
                    if i % self.grid_size[1] == 0:
                        self.agent_P[i, j, i] = Decimal(str(1))
                    else:
                        self.agent_P[i, j, i-1] = Decimal(str(1))

                if not ((i - self.grid_size[1] < 0) or
                        (i + self.grid_size[1] >= self.num_cells) or
                        (i % self.grid_size[1] == self.grid_size[1]-1) or
                        (i % self.grid_size[1] == 0)):

                    # generate stochastic dynamics for weak actions 
                    if j in [0, 1, 2, 3]:
                        self.agent_P[i, j, :] = self._generate(i, j)
                        
        # initialize agent and environment
        self.reset(self.init_state)
        self._add_reloads()


    def _generate(self, s, a):
        '''
        Given state and action, generate the transition probability array and
        the reload states for weak actions.
        '''
        # combined actual heading
        actual_x = self.AGENT_V*np.cos(self.THETA_LIST[a]) + self.FLOW_MAG[s,a]*np.cos(self.FLOW_THETA[s, a])
        actual_y = self.AGENT_V*np.sin(self.THETA_LIST[a]) + self.FLOW_MAG[s,a]*np.sin(self.FLOW_THETA[s, a])
        actual_theta = np.arctan2(actual_y, actual_x)
        rv = norm(actual_theta, self.HEADING_SD)

        # calculate TP
        tp = np.zeros(self.num_cells)
        tp[s-self.grid_size[1]] = p_n = Decimal(str(round(rv.cdf(3*np.pi/4) - rv.cdf(np.pi/4), 3)))
        tp[s+self.grid_size[1]] = p_s = Decimal(str(round(rv.cdf(-np.pi/4) - rv.cdf(-3*np.pi/4), 3)))
        tp[s+1] = p_e = Decimal(str(round(rv.cdf(np.pi/4) - rv.cdf(-np.pi/4), 3)))
        diff = Decimal(str(round(1 - (p_n + p_s + p_e),3)))
        if p_n + p_s + p_e > Decimal(1):
            tp[s-1] = Decimal(str(0))
            tp[s+1] = Decimal(str(round(Decimal(tp[s+1]) - diff,3))) 
        else:
            tp[s-1] = diff
        return tp
    
    
    def update_strategy(self, strategy):
        '''
        Update the strategy attribute to the given stategy
        '''
        self.strategy = strategy
    
            
    def pick_action(self, strategy=None, state=None, energy=None):
        '''
        Given strategy, current state, and current energy level pick the next
        action of the agent
        '''
        if state==None:
            state = self.position
        if energy==None:
            energy = self.agent_energy
        if strategy==None:
            strategy = self.strategy
            if strategy == None:
                raise Exception('Add the strategy by using env.update_strategy method')
        data_dict = self.strategy[state]
        dict_keys = list(data_dict.keys())
        if len(dict_keys) == 0:
            raise Exception('Strategy does not prescribe a safe action at this state. Increase the capacity of the agent.')
        feasible = []
        for value in dict_keys:
            if value <= energy:
                feasible.append(value)
        action_string = data_dict[max(feasible)]
        agent_action = ast.literal_eval(action_string)[1]
        return agent_action
        

    def step(self, action, do_render=0):
        '''
        Given action, function step updates the state of the agent and returns
        the resultant state, cap, and other info
        '''

        # initialize variables
        done = 0
        if self.agent_energy < self.saction_cost:
            raise Exception('Warning: energy too low to take strong action.')
        
        # verify action input and decide action
        if isinstance(action, int) and action in list(range(8)):
            self.position = np.random.choice(self.num_cells,
                                             p=self.agent_P[self.position, action, :])
        else:
            raise Exception('Input action is not a valid action')

        # update energy
        if action in [0,1,2,3]:
            self.agent_energy -= self.waction_cost
        else:
            self.agent_energy -= self.saction_cost
        if self.is_reload[self.position]:
            self.agent_energy = self.agent_capacity
        
        # energy exhaustion check
        if self.agent_energy <= 1 and self.agent_energy >= 0:
            done = 1
       
        # update other terms
        self.states_history.append(self.position)
        self.actions_history.append(action)
        self.time_steps += 1
        
        # render and return
        if do_render == 1:
            self.render_grid()
        info = (self.position, action, self.agent_energy, done)
        return info
    

    def reset(self, init_state=None, reset_energy=None):

        # random initial position
        if init_state == None:
            self.position = np.random.randint(self.num_cells)
        else:
            if 0 <= init_state < self.num_cells and isinstance(init_state, int):
                self.position = init_state
            else:
                raise Exception(
                    'states should be integers between 0 and num_cells. Check init_state again')
                
        if reset_energy == None:
            self.agent_energy = self.agent_capacity
        else:
            self.agent_energy = reset_energy
        self.states_history = [self.position]
        self.actions_history = []
        self.targets_history = []
        self.time_steps = 0
    

    def _add_reloads(self):    
        """
        Internal method to assign reload states.
        """        
        for state in self.reload_list:
            self.is_reload[state] = 1
            
        
    def _get_dist(self, mdp, state, action):
        """
        Internal method that returns a dictionary of states with nonzero probabilities for
        a given state and action
        """
        dist = dict()
        agent_dist = self.agent_P[state, action, :]
        agent_posstates = [] 
        for i in agent_dist.nonzero()[0]:
            agent_posstates.append(i)
        for i in list(agent_posstates):
            prob = agent_dist[i]
            dist[i] = prob
        return dist
    
    
    def create_consmdp(self):
        """
        Method to export the UUV gridworld and target states into a pre-defined
        standard consMDP form. Returns MDP object and the target set.
        """
        mdp = consMDP.ConsMDP()
    
        # Add states
        for i in range(self.num_cells):
            if self.is_reload[i]:
                mdp.new_state(True, str(i))  # (reload, label)
            else:
                mdp.new_state(False, str(i))
                
        # List all possible states and actions
        states_list = []
        for i in range(self.num_cells):
                states_list.append(i)
            
        actions_list = []
        for i in range(self.num_actions):
                actions_list.append(i)        

        # Extract and add actions to the MDP
        actions = dict()
        for state in states_list:
            for action in actions_list:
                action_label = str([state,action])
                dist = self._get_dist(mdp, state, action)
                if action in [0,1,2,3]: # weak action
                    actions[action_label] = {"from":str(state), "cons":self.waction_cost, "dist":dist}
                else: # strong action
                    actions[action_label] = {"from":str(state), "cons":self.saction_cost, "dist":dist}
                
        for label, act in actions.items():
            fr = mdp.state_with_name(act["from"])
            mdp.add_action(fr, act["dist"], label, act["cons"])
        
        self.consmdp = mdp
    
    
    def get_consmdp(self):
        """
        Method that returns the consMDP object and target set that already exists
        or generates it if it does not exist
        """
        
        if self.consmdp == None:
            self.create_consmdp()
            return (self.consmdp, self.target_list)
        else:
            return (self.consmdp, self.target_list)

    
    def _states_to_colors(self):
        '''
        Assign colors to the cells based on their current identity.
        '''
        # 0: light blue; 1 : light gray; 2 : dark gray; 3 : green; 4 : orange; 5: dark blue
        data = np.zeros([self.grid_size[0],self.grid_size[1],3],dtype=np.float32)
        data[:] = COLORS[0] # baseline state color
        for cell in self.states_history:
            data[cell//self.grid_size[1], cell%self.grid_size[1]] = COLORS[2] # history
        for cell in self.target_list:
            data[cell//self.grid_size[1], cell%self.grid_size[1]] = COLORS[3] # targets
        for cell in self.reload_list:
            data[cell//self.grid_size[1], cell%self.grid_size[1]] = COLORS[4] # reloads
        data[self.position//self.grid_size[1], self.position%self.grid_size[1]] = COLORS[5] # current state
        return data


    def render_grid(self):
        
        img_data = self._states_to_colors()
        fig, ax = plt.subplots()
        ax.axis('off')
        plt.title("Agent Energy: {}, Time Steps: {}".format(self.agent_energy, self.time_steps))
        plt.imshow(img_data) 
        plt.show()


    def animate_strategy(self, strategy=None, num_steps=100, interval=100):
        '''
        Run the strategy and animate for num_steps no.of time steps
        '''
        
        if strategy != None:
            self.strategy = strategy
        self.reset(self.init_state)
        fig = plt.figure()
        ax = fig.gca()
        ax.axis('off')
        im = plt.imshow(self._states_to_colors(), animated=True)
        plt.close()
        
        def updatefig(frame_count):
            
            if frame_count == 0:
                im.set_array(self._states_to_colors())
                ax.set_title("Agent Energy: {}, Time Steps: {}".format(self.agent_energy, self.time_steps))
                return im
            self.step(self.pick_action())
            im.set_array(self._states_to_colors())
            ax.set_title("Agent Energy: {}, Time Steps: {}".format(self.agent_energy, self.time_steps))
            return im
        return animation.FuncAnimation(fig, updatefig, frames=num_steps, interval=interval)
        
        
    def _states_to_colors_dev(self):
        '''
        Assign colors to the cells based on their current identity.
        '''
        # 0: light blue; 1 : light gray; 2 : dark gray; 3 : green; 4 : orange; 5: dark blue
        data = np.zeros([self.grid_size[0],self.grid_size[1],3],dtype=np.float32)
        data[:] = COLORS[0] # baseline state color
        
        # separate strong and weak outcomes
        states_strongoutcome = []
        states_weakoutcome = []
        
        for index, action in enumerate(self.actions_history):
            if action in [0,1,2,3]:
                states_weakoutcome.append(self.states_history[index+1])
            elif action in [4,5,6,7]:
                states_strongoutcome.append(self.states_history[index+1])
        
        for cell in states_strongoutcome:
            data[cell//self.grid_size[1], cell%self.grid_size[1]] = COLORS[2] # strong outcomes
        for cell in states_weakoutcome:
            data[cell//self.grid_size[1], cell%self.grid_size[1]] = COLORS[1] # weak outcomes
        for cell in self.target_list:
            data[cell//self.grid_size[1], cell%self.grid_size[1]] = COLORS[3] # targets
        for cell in self.reload_list:
            data[cell//self.grid_size[1], cell%self.grid_size[1]] = COLORS[4] # reloads
        data[self.position//self.grid_size[1], self.position%self.grid_size[1]] = COLORS[5] # current state
        return data
    
    def animate_strategy_dev(self, strategy=None, num_steps=100, interval=100):
        '''
        Run the strategy and animate for num_steps no.of time steps while highlighting
        outcomes of strong and weak actions separately
        '''
        
        if strategy != None:
            self.strategy = strategy
        self.reset(self.init_state)
        fig = plt.figure()
        ax = fig.gca()
        ax.axis('off')
        im = plt.imshow(self._states_to_colors_dev(), animated=True)
        plt.close()
        
        def updatefig(frame_count):
            if frame_count == 0:
                im.set_array(self._states_to_colors())
                ax.set_title("Agent Energy: {}, Time Steps: {}".format(self.agent_energy, self.time_steps))
                return im            
            self.step(self.pick_action())
            im.set_array(self._states_to_colors_dev())
            ax.set_title("Agent Energy: {}, Time Steps: {}".format(self.agent_energy, self.time_steps))
            return im
        return animation.FuncAnimation(fig, updatefig, frames=num_steps, interval=interval)


    def _repr_png_(self):
        '''
        Show graphical representation in notebooks.
        '''
        return self.render_grid()


from setuptools import setup
import pathlib

from fimdpenv import __version__ as version

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='fimdpenv',
    version=version,
    description='Simulation environments for FiMDP Package',
    long_description=README,
    long_description_content_type="text/markdown",
    keywords='simulation-environment, grid-world, agent, cmdp, uuv-dynamics, street-network, formal-methods',
    url='https://github.com/pthangeda/FiMDPEnv',
    author="Pranay Thangeda",
    author_email="contact@prny.me",
    license="MIT",
    python_requires=">=3.6.0",
    packages=['fimdpenv'],
    install_requires=[
        'fimdp>=1.0.2',
        'numpy>=1.18.5',
        'matplotlib>=3.2.2',
        'scipy>=1.5.0',
        'ipython>=7.16.1',
    ]
)



"""
Multi-agent grid world testbed that simulates the motion of a rover and a helicopter 
operating on Martian surface with obstacles.
"""

import numpy as np
from numpy import linalg as LA
from decimal import Decimal
from fimdp import consMDP

class MarsEnv:
    """Class that models a consumption Markov Decision Process with two agents following different dynamics. The 
    consumption MDP models the motion of a Martian rover and a aerial vehicle where the aerial vehicle can only reload
    when it occupies the same state as the rover. The aerial vehicle is referred to as agent.

    :param grid_size: Positive integer that denotes the size of the 2D grid to be generated. 
    :type grid_size: int
    :param agent_capacity: Non negative number denoting the energy capacity of the agent
    :type agent_capacity: float
    :param agent_actioncost: Positive number that denotes the energy consumed by the agent for any action other than stay, defaults to 1.
    :type agent_actioncost: float
    :param agent_staycost: Positive number that denotes the energy consumed by the agent for the stay action, defaults to 1.
    :type agent_staycost: float
    :param init_state: A tuple of length two denoting the initial state of the rover and the agent, defaults to None
    :type init_state: tuple, optional
    """

    def __init__(self, grid_size, agent_capacity, agent_actioncost=1, agent_staycost=1, init_state=None):
        """Constructor method
        """
        # User inputs and other required constants
        self.grid_size = grid_size
        self.agent_actioncost = agent_actioncost
        self.agent_staycost = agent_staycost
        self.agent_capacity = agent_capacity
        self.count = 0
        self.unreachable_cells = []
        if not isinstance(grid_size, int):
            raise TypeError("Grid size grid_size must be an integer greater than 0")

        # initialize required variables
        self.num_cells = grid_size**2
        self.num_actions = 5
        self.state = init_state
        self.agent_energy = self.agent_capacity

        # initialize grid and transition probabilities
        self.rover_P = np.zeros([self.num_cells, self.num_actions, self.num_cells])
        self.agent_P = np.zeros([self.num_cells, self.num_actions, self.num_cells])

        # generate dynamics for rover and agent
        # agent dynamics - deterministic. Rover dynamics - stochastic
        for i in range(self.num_cells):
            for j in range(self.num_actions):

                # north - action 0
                if j == 0:
                    if i - self.grid_size < 0:
                        self.rover_P[i,j,i] = 1
                        self.agent_P[i,j,i] = 1
                    else:
                        self.rover_P[i,j,i-self.grid_size] = 1
                        self.agent_P[i,j,i-self.grid_size] = 1

                # south - action 1
                elif j == 1:
                    if i + self.grid_size >= self.num_cells:
                        self.rover_P[i,j,i] = 1
                        self.agent_P[i,j,i] = 1
                    else:
                        self.rover_P[i,j,i+self.grid_size] = 1
                        self.agent_P[i,j,i+self.grid_size] = 1

                # east - action 2
                elif j == 2:
                    if i%self.grid_size == self.grid_size-1:
                        self.rover_P[i,j,i] = 1
                        self.agent_P[i,j,i] = 1
                    else:
                        self.rover_P[i,j,i+1] = 1
                        self.agent_P[i,j,i+1] = 1

                # west - action 3
                elif j == 3:
                    if i%self.grid_size == 0:
                        self.rover_P[i,j,i] = 1
                        self.agent_P[i,j,i] = 1
                    else:
                        self.rover_P[i,j,i-1] = 1
                        self.agent_P[i,j,i-1] = 1

                # stay - action 4
                elif j == 4:
                    self.rover_P[i,j,i] = 1
                    self.agent_P[i,j,i] = 1

                # rover stochastic dynamics sampled from dirichlet distribution
                if not ((i - self.grid_size < 0) or \
                        (i + self.grid_size >= self.num_cells) or \
                        (i%self.grid_size == self.grid_size-1) or \
                        (i%self.grid_size == 0)):

                    eps = 1e-10;
                    alpha = eps*np.ones(self.num_cells)
                    alpha.put([i, i-self.grid_size, i+self.grid_size,
                               i+1, i-1],2*np.ones(self.num_actions))
                    outcomes = [i-self.grid_size,i+self.grid_size,i+1,i-1,i]
                    alpha.put(outcomes[j],10)
                    if j == 4:
                        pass
                    else:
                        self.rover_P[i,j,:] = np.random.dirichlet(alpha, size=1)

        # generate unreachable cells
        self._unreachablestates()

        # initialize agent and rover positions
        self.reset(init_state)

    def _unreachablestates(self):
        """
        Internal method that generates a set of cells are made unreachable to the rover i.e. rover dynamics 
        are modified to make sure rover can never enter these cells irrespective of the action it takes.
        """

        if self.grid_size < 5:
            raise Exception("Grid size too small to generate unreachable cells; use a grid size no smaller than 5.")
        else:

            # declare unreachable cells
            centercell = [(self.grid_size-1)//2, (self.grid_size-1)//2]
            radius = (self.grid_size)//3

            for i in range(self.grid_size):
                boundary = []
                unreachable = []
                for i in range(self.num_cells):
                    # boundary cells
                    if LA.norm([i//self.grid_size-centercell[0], i%self.grid_size-centercell[1]]) <= radius:
                        boundary.append(i)
                    # unreachable cells
                    if LA.norm([i//self.grid_size-centercell[0], i%self.grid_size-centercell[1]]) <= radius-1:
                        unreachable.append(i)

            # change dynamics of boundary cells
            for cell in boundary:

                # new transiton array
                array = np.zeros(self.num_cells)
                array.put(cell,1)

                # north - 0
                if cell - self.grid_size in unreachable:
                    self.rover_P[cell, 0,:] = array
                    # Other actions
                    for i in [1,2,3]:
                        self.rover_P[cell, i, cell-self.grid_size] = 0
                    for i in [0,1,2,3]:
                        self.rover_P[cell, i,:] = self.rover_P[cell, i,:]/np.sum(self.rover_P[cell, i,:])

                # south - 1
                if cell + self.grid_size in unreachable:
                    self.rover_P[cell, 1,:] = array
                    # Other actions
                    for i in [0,2,3]:
                        self.rover_P[cell, i, cell+self.grid_size] = 0
                    for i in [0,1,2,3]:
                        self.rover_P[cell, i,:] = self.rover_P[cell, i,:]/np.sum(self.rover_P[cell, i,:])

                # east - 2
                if cell + 1 in unreachable:
                    self.rover_P[cell, 2,:] = array
                    # Other actions
                    for i in [0,1,3]:
                        self.rover_P[cell, i, cell+1] = 0
                    for i in [0,1,2,3]:
                        self.rover_P[cell, i,:] = self.rover_P[cell, i,:]/np.sum(self.rover_P[cell, i,:])

                # west - 3
                if cell - 1 in unreachable:
                    self.rover_P[cell, 3,:] = array
                    # Other actions
                    for i in [0,1,2]:
                        self.rover_P[cell, i, cell-1] = 0
                    for i in [0,1,2,3]:
                        self.rover_P[cell, i,:] = self.rover_P[cell, i,:]/np.sum(self.rover_P[cell, i,:])

            # change dynamics of unreachable cells
            for cell in unreachable:
                self.rover_P[cell,:,:] = np.zeros([self.num_actions, self.num_cells])
                self.rover_P[cell,:,cell] = np.ones(self.num_actions)

            # update unreachable cells
            self.unreachable_cells = unreachable

    def step(self, action):
        """
        Method that given action [agent action, rover action], function step updates the
        position of the rover and agent in the grid and also the energy of the
        agent. 
        
        Notes
        -----
        Stochastic actions inputs are also allowed.
        """

        # initialize variables
        done = 0
        previous_position = self.position

        # verify action input and decide action
        if np.shape(action) == (2,):
            pass
        elif np.shape(action) == (2,self.num_actions):
            if all(sum(act) == 1 for act in action):
                agent_act = np.random.choice(self.num_actions, p=action[0])
                rover_act = np.random.choice(self.num_actions, p=action[1])
                action = [agent_act, rover_act]
            else:
                raise Exception('Action probabilities do not add up to 1')
        else:
            raise Exception('Action dimensions are wrong. Please input action in one of the two suggested dimensions')

        ## update position
        
        # agent position
        self.position[0] = np.random.choice(self.num_cells,
                  p=self.agent_P[self.position[0],action[0],:])
        # rover position
        self.position[1] = np.random.choice(self.num_cells,
                  p=self.rover_P[self.position[1],action[1],:])

        ## update energy
        if previous_position[0] == self.position[0]:
            self.agent_energy -= self.agent_staycost
        else:
            self.agent_energy -= self.agent_actioncost
        # reload if feasible
        if self.position[0] == self.position[1]:
            self.agent_energy = self.agent_capacity
        ## out of energy flag
        if self.agent_energy == 0:
            done = 1
        elif self.agent_energy < 0:
            raise Exception('The energy of the agent cannot be negative.')

        # data export structure
        info = (tuple(self.position), tuple(action), self.agent_energy, done)
        return info

    def reset(self, init_state=None):
        """
        Method that resets the position of the rover and the agent and also the energy of the agent.
        It is ensured that the agent and the rover and reset to a feasible state. 
        """

        # random initial position
        if init_state == None:
            self.position = list(np.random.randint(self.num_cells,size=2))
        elif np.shape(init_state) == (2,):
            if all(0 <= state < self.num_cells for state in init_state) and \
            all(isinstance(state, int) for state in init_state):
                self.position = list(init_state)
            else:
                raise Exception('states should be integers between 0 and num_cells. Check init_state again')
        else:
            raise Exception('init_state should should be a 1X2 list')

        # override initial state if rover is in unreachable positions
        if self.position[1] in self.unreachable_cells:
            while True:
                self.position = list(np.random.randint(self.num_cells,size=2))
                if self.position[1] not in self.unreachable_cells:
                    break

        # reset agent energy
        self.agent_energy = self.agent_capacity

    def _get_targets(self, mdp):    
        """
        Internal method that generates target states that are well distributed over reachable
        and unreachable states and are a pre-defined proportion of the total no.of states. 
        """        
        
        target_prop = 0.2 # % of all states that will be targets
        
        # randomly select states
        T = set()
        state_list = mdp.names
        for state in state_list:
            if np.random.rand() <= target_prop:
                T.add(mdp.state_with_name(state))
        return T
        
        
    def _get_dist(self, mdp, state, action):
        """
        Internal method that returns a dictionary of states with nonzero probabilities for
        a given state and action
        """
        
        # Initialize
        dist = dict()
        agent_dist = self.agent_P[state[0], action[0], :]
        rover_dist = self.rover_P[state[1], action[1], :]
        agent_posstates = agent_dist.nonzero()
        rover_posstates = rover_dist.nonzero()
            
        # Store non-zero transition probabilities
        for i in list(agent_posstates[0]):
            for j in list(rover_posstates[0]):
                state = str((i,j))
                prob = agent_dist[i]*rover_dist[j]
                sid = mdp.state_with_name(str(state))
                dist[sid] = prob
    
        # Normalize to ensure probabilities add up exactly to 1
        s = sum(dist.values())
        for key in dist:
            dist[key] = round(Decimal(str(dist[key]/s)),4)
        key = list(dist.keys())[-1]
        dist[key] = Decimal(1) - sum(list(dist.values())[:-1])
        return dist
    
    def get_mdp_targets(self):
        """
        Method to export the martian gridworld and target states into a pre-defined
        standard form. Returns MDP object and the target set.
        """
        mdp = consMDP.ConsMDP()
    
        # Add states
        for i in range(self.num_cells):
            for j in range(self.num_cells):
                if i == j:
                    mdp.new_state(True, str((i, j)))  # (reload, label)
                else:
                    mdp.new_state(False, str((i, j)))
                
        # List all possible states and actions
        states_list = []
        for i in range(self.num_cells):
            for j in range(self.num_cells):
                states_list.append((i,j))
            
        actions_list = []
        for i in range(self.num_actions):
            for j in range(self.num_actions):
                actions_list.append((i,j))        

        # Extract and add actions to the MDP
        actions = dict()
        for state in states_list:
            for action in actions_list:
                action_label = str([state, action])
                dist = self._get_dist(mdp, state, action)
                if action[0] == action[1] == 4: # stay
                    actions[action_label] = {"from":str(state), "cons":self.agent_staycost, "dist":dist}
                else:
                    actions[action_label] = {"from":str(state), "cons":self.agent_actioncost, "dist":dist}
                
        for label, act in actions.items():
            fr = mdp.state_with_name(act["from"])
            mdp.add_action(fr, act["dist"], label, act["cons"])

        # Get targets
        target_set = self._get_targets(mdp)
    
        return (mdp, target_set)
